from typing import MutableMapping, Union, List, Any, Dict


class DotList(list):
    def __getitem__(self, index):
        value = super().__getitem__(index)
        if isinstance(value, dict):
            return Dot(value)
        return value

    def flatten(self, allow: List[str] = None) -> List[Dict[str, Any]]:
        result = [item.flat() for item in self]
        if allow is not None:
            result = [
                {k: v for k, v in d.items() if k.startswith(tuple(allow))}
                for d in result
            ]
            # remove empty dicts
            result = [d for d in result if d]
        return result

    def merge(self) -> dict:
        list = self.flatten()
        result = {}
        for item in list:
            result.update(item)
        return result


class Dot(dict):

    def __getattr__(self, item) -> Union['Dot', DotList]:
        value = self[item]
        if isinstance(value, dict):
            return Dot(value)
        elif isinstance(value, list):
            return DotList(value)
        return value



from durable_dot_dict.dot import DotList
from durable_dot_dict.dotdict import DotDict


def test_dot_list1():
    d1 = DotDict({
        "a": {
            "b": 1,
            "c": {
                "v1": [{"d": 2}],
                "v2": 1
            }
        }
    })
    d2 = DotDict({
        "a": {
            "c": [{"d": 123}]
        }
    })
    dot_list = DotList([d1, d2])
    flat = dot_list.flatten()
    assert isinstance(flat[0], dict)
    merged = dot_list.merge()
    # Data from d1 should remain
    assert merged['a.b'] == d1['a.b']
    assert merged['a.c.v1'] == d1['a.c.v1']
    assert merged['a.c.v2'] == d1['a.c.v2']

    # Added new for d2
    assert merged['a.c'] == d2['a.c']


def test_dot_list2():
    d1 = DotDict({
        "a": {
            "b": 1,
            "c": {
                "v1": [{"d": 2}],
                "v2": 1
            }
        }
    })
    d2 = DotDict({
        "a": {
            "c": {"d": 123, "v2": 123}
        }
    })
    dot_list = DotList([d1, d2])
    flat = dot_list.flatten()
    assert isinstance(flat[0], dict)
    merged = dot_list.merge()
    # Data from d1 should remain
    assert merged['a.b'] == d1['a.b']
    assert merged['a.c.v1'] == d1['a.c.v1']

    # Added new for d2
    assert merged['a.c.d'] == d2['a.c.d']
    assert merged['a.c.v2'] == d2['a.c.v2']


def test_dot_list_flatten():
    d1 = DotDict({
        "a": {
            "b": 1,
            "c": {
                "v1": [{"d": 2}],
                "v2": 1
            }
        }
    })
    d2 = DotDict({
        "a": {
            "c": {"d": 123, "v2": 123}
        }
    })
    dot_list = DotList([d1, d2])
    flat = dot_list.flatten(allow =['a.c.v2'])
    assert flat == [{'a.c.v2': 1}, {'a.c.v2': 123}]


from typing import Any, Callable, List, Iterable, Type, Optional, Union
from typing import Generator, Iterator, TypeVar
from durable_dot_dict.dotdict import DotDict

T = TypeVar("T", bound=DotDict)


def first(*args: Callable[[], Any]) -> Any:
    for func in args:
        try:
            return func()
        except Exception:
            continue
    raise ValueError("All provided callables failed.")


def first_not_none(*args: Callable[[], Any]) -> Any:
    for func in args:
        try:
            result = func()
            if result is None:
                continue
            return result
        except Exception:
            continue
    raise ValueError("All provided callables failed.")


class DotDictStream:
    def __init__(self, collection: Iterable[T]):
        self.collection = collection

    def __iter__(self):
        yield from self.collection

    def __rshift__(self, other) -> 'DotDictStream':
        return DotDictStream(item >> other for item in self.collection)

    def __lshift__(self, other) -> 'DotDictStream':
        return DotDictStream(item << other for item in self.collection)

    def list(self, cast_to: Optional[Type[T]] = None) -> List[T]:
        if cast_to is None:
            return list(self.collection)
        return [item.cast_to(cast_to) for item in self.collection]

    def first(self, cast_to: Optional[Type[T]] = None) -> Optional[Union[T, DotDict]]:
        try:
            item = next(self.collection)

            if cast_to is None:
                return item
            return item.cast_to(cast_to)
        except StopIteration:
            return None

    def remap(self, mapping) -> Generator[DotDict, None, None]:
        for item in self.collection:
            yield item >> mapping
